/* $Id$ */

/* hide.c -- functions to hide data in other files --
 * Copyright (C) 2000 Synack Systems Corporation
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <png.h>
#include <errno.h>
#include <setjmp.h>
#include "common.h"
#include "hide.h"
#include "blowfish.h"

#include "setup.h"

int percentage = 10;
BLOWFISH_context bf;

int hide(byte *data, u32 datalen, byte *noise, u32 noiselen, u32 noiseskip, byte noiseunitsz, u32 randnum)
{
    u32 i, k, ratio, pos, next;
    HIDE_HEADER header;
    byte *p, bit;
    u32 hidespots[HIDE_HEADER_SZ][8];
    int h1, h2;
    int status=0;

    /* make a header that the seek program can use to figure out what to do */
    memset(&header, 0, sizeof(header));
    header.datalen = htonl(datalen);
    header.percent = percentage;
    header.ver = 1;


    /* hide the header in such a way that we can find it knowing only the key */
    p = (byte*)&header;
    for(i=0;i<sizeof(header);i++) {
        for(k=0;k<8;k++) {

            /* figure out where to store the next bit.. we have to advance pos
             * by at least 1 and make sure that pos doesn't land on a byte
             * that we shouldn't use
             */
            do {
                bfencrypt(&bf, &randnum, &randnum);
                next = randnum % noiselen;
            } while(!(next%noiseskip));

            hidespots[i][k] = pos = next;

            bit = (p[i] >> k) & 0x01;
            if(bit) noise[pos] |= 0x01;
            else noise[pos] &= 0xFE;

            // printf("storing bit %d at pos %d or %d\n", bit, pos, noiselen);
        }
    }


    /* figure out how often we need to store a bit to fit everything in the
     * noise and keep the data pretty evenly distributed throughout the file
     */
    ratio = noiselen / ((datalen+sizeof(header))*8);

    /* multiplying the ration by 1.7 seems to be a safe move -- but it may not
     * always work!
     *
     * what we should really do is a test run of the next for loop and find
     * out what the largest ratio we can use is..
     *
     * the larger the ratio the more evenly distributed data will be
     * throughout the noise..
     */
#ifndef COMPAT_098
    ratio *= 1.7;
#endif

    printf("ratio: 1:%d\n", ratio);
    pos = 0;

    p = (byte*)data;
    for(i=0;i<datalen;i++) {
        for(k=0;k<8;k++) {

        tryagain:
            do {
                do {
                    bfencrypt(&bf, &randnum, &randnum);
                    next = randnum % ratio;
                } while(!next);
            } while(!((pos+next)%noiseskip));

            pos+=next;

            if(pos > noiselen) {
                fprintf(stderr, "Sorry, but Refugee ran out of room trying to store your data.\n");
                fprintf(stderr, "This error should be eliminated in future versions.\n");
                status=1; goto cleanup;
            }

            /* make sure we aren't going to overwrite a bit we already
               stored */
            for(h1=0;h1<HIDE_HEADER_SZ;h1++)
                for(h2=0;h2<8;h2++)
                    if(hidespots[h1][h2] == pos)
                        goto tryagain;

            /* store our data bit */
            bit = (p[i] >> k) & 0x01;
            if(bit) noise[pos] |= 0x01;
            else noise[pos] &= 0xFE;

            //printf("storing bit %d at pos %d of %d\n", bit, pos, noiselen);
        }
    }

    printf("%d/%d\n", pos, noiselen);

cleanup:

    memset(&header, 0, sizeof(header));
    memset(&hidespots, 0, sizeof(hidespots));

    return status;
}



int png_hide(char *fname, void *data, u32 datalen, u32 key) {
    FILE *in=NULL, *out=NULL;
    int status=0, sig_read, i;
    byte header[4];
    png_structp png_ptr=NULL, png_write_ptr=NULL;
    png_infop info_ptr=NULL;
    png_uint_32 width, height;
    int depth, color_type, interlace_type;
    png_bytep *row_pointers=NULL;
    byte *image_data=NULL;
    u32 maxfit;
    u32 rowbytes, channels;

    if((in = fopen(fname, "rb")) == NULL) {
        fprintf(stderr, "couldn't open %s for reading.\n", fname);
        status=1; goto cleanup;
    }

    sig_read = fread(header, 1, 4, in);
    if(png_sig_cmp(header, 0, sig_read)) {
        fprintf(stderr, "%s is not a valid PNG file.\n", fname);
        status=1; goto cleanup;
    }

    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if(png_ptr == NULL) {
        fprintf(stderr, "unable to initialize PNG library (1)\n");
        status=1; goto cleanup;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if(info_ptr == NULL) {
        fprintf(stderr, "unable to initialize PNG library (2)\n");
        status=1; goto cleanup;
    }

    if(setjmp(png_ptr->jmpbuf)) {
        fprintf(stderr, "an unrecoverable libpng error has occurred.\n");
        status=1; goto cleanup;
    }

    png_init_io(png_ptr, in);
    png_set_sig_bytes(png_ptr, sig_read);
    png_read_info(png_ptr, info_ptr);
    png_get_IHDR(png_ptr, info_ptr, &width, &height, &depth, &color_type,
                 &interlace_type, NULL, NULL);

    /* maxfit is the max nunmber of bytes that can be stored in an image */
    /* we only work with the RGB components of the image and skip the alpha */
    /* only use percentage% of the pixels */
    /* takes 8 pixels to hold one byte */
    maxfit = (width*height*3) / percentage / 8;
    printf("%s is %ldx%ldx%d and can hold %d bytes of data.\n", fname, width,
           height, depth, maxfit);
    if(maxfit < datalen+sizeof(HIDE_HEADER)) {
        fprintf(stderr, "Refugee can not hide your data.\n%s is not big enough!\n", fname);
        status=66; goto cleanup;
    }

    if(color_type == PNG_COLOR_TYPE_GRAY) {
        fprintf(stderr, "You need to use a color PNG file; %s is B&W.\n", fname);
        status=1; goto cleanup;
    }

    /* we want to always work with 4 channels even though we're only going to
     * use 3 of them -- why?  to rest my weary brain!
     */
    if(color_type == PNG_COLOR_TYPE_PALETTE)
        png_set_expand(png_ptr);

    png_read_update_info(png_ptr, info_ptr);

    rowbytes = png_get_rowbytes(png_ptr, info_ptr);
    channels = png_get_channels(png_ptr, info_ptr);

    if((image_data = malloc(rowbytes*height)) == NULL) {
        fprintf(stderr, "malloc failed.\n");
        status=2; goto cleanup;
    }

    row_pointers = (png_bytepp)malloc(height*sizeof(png_bytep));
    if(!row_pointers) {
        fprintf(stderr, "malloc failed.\n");
        status=2; goto cleanup;
    }

    for(i=0;i<height;i++)
        row_pointers[i] = image_data + i*rowbytes;

    png_read_image(png_ptr, row_pointers);

    //free(row_pointers);
    //row_pointers = NULL;

    png_read_end(png_ptr, info_ptr);
    fclose(in); in=NULL;

    status=hide(data, datalen, image_data, rowbytes*height, 4, depth, key);
    if(status) goto cleanup;

    printf("writing new png file... "); fflush(stdout);

    png_write_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL,
                                            NULL);
    if(!png_write_ptr) {
        fprintf(stderr, "png_create_write_struct() failed.\n");
        status=1; goto cleanup;
    }

    if(setjmp(png_write_ptr->jmpbuf)) {
        fprintf(stderr, "an unrecoverable libpng error has occurred.\n");
        status=1; goto cleanup;
    }

    if((out = fopen(fname, "wb")) == NULL) {
        fprintf(stderr, "couldn't open %s for writing.\n", fname);
        status=1; goto cleanup;
    }

    png_init_io(png_write_ptr, out);
    png_set_compression_level(png_write_ptr, Z_BEST_COMPRESSION);
    png_set_compression_strategy(png_write_ptr, Z_DEFAULT_STRATEGY);
    png_set_compression_mem_level(png_ptr, 8);
    png_set_compression_window_bits(png_ptr, 15);
    png_set_compression_method(png_ptr, 8);

    png_write_info(png_write_ptr, info_ptr);
    png_set_packing(png_write_ptr);
    png_write_image(png_write_ptr, row_pointers);
    png_write_end(png_write_ptr, NULL);

    printf("done.\n");

cleanup:
    if(row_pointers) free(row_pointers);
    if(image_data) {
        memset(image_data, 0, rowbytes*height);
        free(image_data);
    }

    if(png_ptr)
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    if(png_write_ptr)
        png_destroy_write_struct(&png_write_ptr, NULL);

    memset(fname, 0, strlen(fname));

    datalen=0;
    key=0;

    if(in) fclose(in);
    if(out) fclose(out);

    return status;
}

int main(int argc, char **argv) {
    int status=0;
    struct stat st;
    int in=0, i;
    byte *data=NULL;
    byte *bfkey=NULL;
    u32 keylen, *keyptr,randnum;


    if(argc != 3) {
        fprintf(stderr, "usage: %s data_filename png_filename\n", argv[0]);
        fprintf(stderr, "Refugee will hide the contents of data_filename inside the image png_filename\n");
        exit(1);
    }

    nocore();

    in = open(argv[1], O_RDONLY);
    fstat(in, &st);
    data = malloc(st.st_size);
    if(!data) {
        fprintf(stderr, "malloc failed.\n");
        status=1; goto cleanup;
    }
    read(in, data, st.st_size);
    close(in);
    in=0;

    bfkey = getbfkey(&keylen);
    if(!bfkey) {
        fprintf(stderr, "Couldn't get a key.  Run key.\n");
        status=1; goto cleanup;
    }
    if(bf_setkey(&bf, bfkey, keylen) != 0) {
        fprintf(stderr, "error setting blowfish key.\n");
        status=1; goto cleanup;
    }

    randnum=0;
    keyptr = (u32*)bfkey;
    for(i=0;i<keylen/4;i++)
        randnum ^= ntohl(*(keyptr+i));

    status = png_hide(argv[2], data, st.st_size, randnum);

cleanup:
    if(data) {
        memset(data, 0, st.st_size);
        free(data);
    }

    if(bfkey)
        memset(bfkey, 0, keylen);

    keylen = 0;

    memset(&bf, 0, sizeof(bf));

    if(in) close(in);

    memset(&st, 0, sizeof(st));

    return status;
}
